/* -LICENSE-START-
 ** Copyright (c) 2018 Blackmagic Design
 **
 ** Permission is hereby granted, free of charge, to any person or organization
 ** obtaining a copy of the software and accompanying documentation covered by
 ** this license (the "Software") to use, reproduce, display, distribute,
 ** execute, and transmit the Software, and to prepare derivative works of the
 ** Software, and to permit third-parties to whom the Software is furnished to
 ** do so, all subject to the following:
 **
 ** The copyright notices in the Software and this entire statement, including
 ** the above license grant, this restriction and the following disclaimer,
 ** must be included in all copies of the Software, in whole or in part, and
 ** all derivative works of the Software, unless such copies or derivative
 ** works are solely in the form of machine-executable object code generated by
 ** a source language processor.
 **
 ** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 ** IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 ** FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
 ** SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
 ** FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
 ** ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 ** DEALINGS IN THE SOFTWARE.
 ** -LICENSE-END-
 */

#include "BlackmagicRawAPI.h"

#include <iostream>

#include <CoreServices/CoreServices.h>
#include <ImageIO/ImageIO.h>

#ifdef DEBUG
	#include <cassert>
	#define VERIFY(condition) assert(SUCCEEDED(condition))
#else
	#define VERIFY(condition) condition
#endif

static const BlackmagicRawResourceFormat s_resourceFormat = blackmagicRawResourceFormatRGBAU8;
static const CFStringRef s_outputFileName = CFSTR("/tmp/outputRGBAU8.png");

static void OutputImage(CFStringRef outputFileName, uint32_t width, uint32_t height, uint32_t sizeBytes, void* imageData)
{
	bool success = false;
	const char* outputFileNameAsCString = CFStringGetCStringPtr(outputFileName, kCFStringEncodingMacRoman);

	CFURLRef file = CFURLCreateWithFileSystemPath(kCFAllocatorDefault, outputFileName, kCFURLPOSIXPathStyle, false);
	if (file != nullptr)
	{
		const uint32_t bitsPerComponent	= 8;
		const uint32_t bitsPerPixel		= 32;
		const uint32_t bytesPerRow		= (bitsPerPixel * width) / 8U;

		CGColorSpaceRef space			= CGColorSpaceCreateWithName(kCGColorSpaceSRGB);
		CGBitmapInfo bitmapInfo			= kCGImageAlphaNoneSkipLast | kCGImageByteOrderDefault;
		CGDataProviderRef provider		= CGDataProviderCreateWithData(nullptr, imageData, sizeBytes, nullptr);
		const CGFloat* decode			= nullptr;
		bool shouldInterpolate			= false;
		CGColorRenderingIntent intent	= kCGRenderingIntentDefault;

		CGImageRef imageRef = CGImageCreate(width, height, bitsPerComponent, bitsPerPixel, bytesPerRow, space, bitmapInfo, provider, decode, shouldInterpolate, intent);
		if (imageRef != nullptr)
		{
			CGImageDestinationRef destination = CGImageDestinationCreateWithURL(file, kUTTypePNG, 1, nullptr);
			if (destination)
			{
				CGImageDestinationAddImage(destination, imageRef, nil);
				CGImageDestinationFinalize(destination);

				CFRelease(destination);

				std::cout << "Created " << outputFileNameAsCString << std::endl;
				success = true;
			}

			CGImageRelease(imageRef);
		}

		CGDataProviderRelease(provider);
		CGColorSpaceRelease(space);

		CFRelease(file);
	}

	if (! success)
		std::cerr << "Failed to create " << outputFileNameAsCString << "!" << std::endl;
}

class CameraCodecCallback : public IBlackmagicRawCallback
{
public:
	explicit CameraCodecCallback() = default;
	virtual ~CameraCodecCallback() = default;

	virtual void ReadComplete(IBlackmagicRawJob* readJob, HRESULT result, IBlackmagicRawFrame* frame)
	{
		IBlackmagicRawJob* decodeAndProcessJob = nullptr;

		if (result == S_OK)
			VERIFY(frame->SetResourceFormat(s_resourceFormat));

		if (result == S_OK)
			result = frame->CreateJobDecodeAndProcessFrame(nullptr, nullptr, &decodeAndProcessJob);

		if (result == S_OK)
			result = decodeAndProcessJob->Submit();

		if (result != S_OK)
		{
			if (decodeAndProcessJob)
				decodeAndProcessJob->Release();
		}

		readJob->Release();
	}

	virtual void ProcessComplete(IBlackmagicRawJob* job, HRESULT result, IBlackmagicRawProcessedImage* processedImage)
	{
		unsigned int width = 0;
		unsigned int height = 0;
		unsigned int sizeBytes = 0;
		void* imageData = nullptr;

		if (result == S_OK)
			result = processedImage->GetWidth(&width);

		if (result == S_OK)
			result = processedImage->GetHeight(&height);

		if (result == S_OK)
			result = processedImage->GetResourceSizeBytes(&sizeBytes);

		if (result == S_OK)
			result = processedImage->GetResource(&imageData);

		if (result == S_OK)
			OutputImage(s_outputFileName, width, height, sizeBytes, imageData);

		job->Release();
	}

	virtual void DecodeComplete(IBlackmagicRawJob*, HRESULT) {}
	virtual void TrimProgress(IBlackmagicRawJob*, float) {}
	virtual void TrimComplete(IBlackmagicRawJob*, HRESULT) {}
	virtual void SidecarMetadataParseWarning(IBlackmagicRawClip*, CFStringRef, uint32_t, CFStringRef) {}
	virtual void SidecarMetadataParseError(IBlackmagicRawClip*, CFStringRef, uint32_t, CFStringRef) {}
	virtual void PreparePipelineComplete(void*, HRESULT) {}

	virtual HRESULT STDMETHODCALLTYPE QueryInterface(REFIID, LPVOID*)
	{
		return E_NOTIMPL;
	}

	virtual ULONG STDMETHODCALLTYPE AddRef(void)
	{
		return 0;
	}

	virtual ULONG STDMETHODCALLTYPE Release(void)
	{
		return 0;
	}
};

int main(int argc, const char* argv[])
{
	if (argc > 2)
	{
		std::cerr << "Usage: " << argv[0] << " clipName.braw" << std::endl;
		return 1;
	}

	CFStringRef clipName;
	bool clipNameProvided = argc == 2;
	if (clipNameProvided)
	{
		clipName = CFStringCreateWithCString(NULL, argv[1], kCFStringEncodingUTF8);
	}
	else
	{
		clipName = CFSTR("../../../Media/sample.braw");
	}

	HRESULT result = S_OK;

	IBlackmagicRawFactory* factory = nullptr;
	IBlackmagicRaw* codec = nullptr;
	IBlackmagicRawClip* clip = nullptr;
	IBlackmagicRawJob* readJob = nullptr;
	IBlackmagicRawClipGeometry* geometry = nullptr;

	CameraCodecCallback callback;
	long readTime = 0;

	do
	{
		factory = CreateBlackmagicRawFactoryInstanceFromPath(CFSTR("../../Libraries"));
		if (factory == nullptr)
		{
			std::cerr << "Failed to create IBlackmagicRawFactory!" << std::endl;
			break;
		}

		result = factory->CreateCodec(&codec);
		if (result != S_OK)
		{
			std::cerr << "Failed to create IBlackmagicRaw!" << std::endl;
			break;
		}
		
		result = factory->CreateClipGeometry(&geometry);
		if (result != S_OK)
		{
			std::cerr << "Failed to create IBlackmagicRawClipGeometry!" << std::endl;
			break;
		}
		
		result = geometry->SetAnamorphicRatio(blackmagicRawAnamorphicRatioFromMetadata);
		if (result != S_OK)
		{
			std::cerr << "Failed to set anamorphic ratio!" << std::endl;
			break;
		}
		
		result = geometry->SetRotation(blackmagicRawRotationClockwise90);
		if (result != S_OK)
		{
			std::cerr << "Failed to set rotation!" << std::endl;
			break;
		}
		
		result = geometry->SetFlip(blackmagicRawFlipDisabled);
		if (result != S_OK)
		{
			std::cerr << "Failed to set flip!" << std::endl;
			break;
		}

		result = codec->OpenClipWithGeometry(clipName, geometry, &clip);
		if (result != S_OK)
		{
			std::cerr << "Failed to open IBlackmagicRawClip!" << std::endl;
			break;
		}

		result = codec->SetCallback(&callback);
		if (result != S_OK)
		{
			std::cerr << "Failed to set IBlackmagicRawCallback!" << std::endl;
			break;
		}

		result = clip->CreateJobReadFrame(readTime, &readJob);
		if (result != S_OK)
		{
			std::cerr << "Failed to create IBlackmagicRawJob!" << std::endl;
			break;
		}

		result = readJob->Submit();
		if (result != S_OK)
		{
			// submit has failed, the ReadComplete callback won't be called, release the job here instead
			readJob->Release();
			std::cerr << "Failed to submit IBlackmagicRawJob!" << std::endl;
			break;
		}

		codec->FlushJobs();

	} while(0);
	
	if (geometry)
		geometry->Release();

	if (clip != nullptr)
		clip->Release();

	if (codec != nullptr)
		codec->Release();

	if (factory != nullptr)
		factory->Release();

	CFRelease(clipName);

	return result;
}
