/* -LICENSE-START-
 ** Copyright (c) 2023 Blackmagic Design
 **
 ** Permission is hereby granted, free of charge, to any person or organization
 ** obtaining a copy of the software and accompanying documentation covered by
 ** this license (the "Software") to use, reproduce, display, distribute,
 ** execute, and transmit the Software, and to prepare derivative works of the
 ** Software, and to permit third-parties to whom the Software is furnished to
 ** do so, all subject to the following:
 **
 ** The copyright notices in the Software and this entire statement, including
 ** the above license grant, this restriction and the following disclaimer,
 ** must be included in all copies of the Software, in whole or in part, and
 ** all derivative works of the Software, unless such copies or derivative
 ** works are solely in the form of machine-executable object code generated by
 ** a source language processor.
 **
 ** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 ** IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 ** FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
 ** SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
 ** FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
 ** ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 ** DEALINGS IN THE SOFTWARE.
 ** -LICENSE-END-
 */

#import "ViewController.h"

#include "BlackmagicRawAPI.h"

#import <Metal/Metal.h>
#include <mutex>							// std::mutex, std::unique_lock
#include <thread>							// std::this_thread::sleep_for

#ifdef DEBUG
	#include <cassert>
	#define VERIFY(condition) assert(SUCCEEDED(condition))
#else
	#define VERIFY(condition) condition
#endif

static const BlackmagicRawResourceFormat s_resourceFormat = blackmagicRawResourceFormatRGBAU8;

class BufferManagerFlow2
{
public:
	BufferManagerFlow2(IBlackmagicRawResourceManager* resourceManager, IBlackmagicRawManualDecoderFlow2* manualDecoder, void* post3DLUTBufferGPU, void* context, void* commandQueue, BlackmagicRawResourceType gpuResourceType)
	: m_resourceManager(resourceManager)
	, m_manualDecoder(manualDecoder)
	, m_post3DLUTBufferGPU(post3DLUTBufferGPU)
	, m_context(context)
	, m_commandQueue(commandQueue)
	, m_gpuResourceType(gpuResourceType)
	{
	}

	~BufferManagerFlow2()
	{
		if (m_bitStream != nullptr)
		{
			VERIFY(m_resourceManager->ReleaseResource(nullptr, nullptr, m_bitStream, blackmagicRawResourceTypeBufferCPU));
		}

		if (m_frameState != nullptr)
		{
			VERIFY(m_resourceManager->ReleaseResource(nullptr, nullptr, m_frameState, blackmagicRawResourceTypeBufferCPU));
		}

		if (m_decodedBufferCPU != nullptr)
		{
			VERIFY(m_resourceManager->ReleaseResource(nullptr, nullptr, m_decodedBufferCPU, blackmagicRawResourceTypeBufferCPU));
		}

		if (m_decodedBufferGPU != nullptr)
		{
			VERIFY(m_resourceManager->ReleaseResource(m_context, m_commandQueue, m_decodedBufferGPU, m_gpuResourceType));
		}

		if (m_processedBuffer != nullptr)
		{
			VERIFY(m_resourceManager->ReleaseResource(m_context, m_commandQueue, m_processedBuffer, m_gpuResourceType));
		}
	}

	void populateFrameStateBuffer(IBlackmagicRawFrame* frame)
	{
		uint32_t frameStateSizeBytes;
		VERIFY(m_manualDecoder->GetFrameStateSizeBytes(&frameStateSizeBytes));

		// Allocate the buffer if necessary
		if (frameStateSizeBytes > m_frameStateSizeBytes)
		{
			if (m_frameState != nullptr)
			{
				VERIFY(m_resourceManager->ReleaseResource(nullptr, nullptr, m_frameState, blackmagicRawResourceTypeBufferCPU));
			}

			VERIFY(m_resourceManager->CreateResource(nullptr, nullptr, frameStateSizeBytes, blackmagicRawResourceTypeBufferCPU, blackmagicRawResourceUsageReadCPUWriteCPU, &m_frameState));
			m_frameStateSizeBytes = frameStateSizeBytes;
		}

		// Populate the buffer
		VERIFY(m_manualDecoder->PopulateFrameStateBuffer(frame, nullptr, nullptr, m_frameState, frameStateSizeBytes));
	}

	void createReadJob(IBlackmagicRawClipEx* clipEx, uint64_t frameIndex, IBlackmagicRawJob** jobRead)
	{
		uint32_t bitStreamSizeBytes = 0;
		VERIFY(clipEx->GetBitStreamSizeBytes(frameIndex, &bitStreamSizeBytes));

		// Allocate the buffer if necessary
		if (bitStreamSizeBytes > m_bitStreamSizeBytes)
		{
			if (m_bitStream != nullptr)
			{
				VERIFY(m_resourceManager->ReleaseResource(nullptr, nullptr, m_bitStream, blackmagicRawResourceTypeBufferCPU));
			}

			VERIFY(m_resourceManager->CreateResource(nullptr, nullptr, bitStreamSizeBytes, blackmagicRawResourceTypeBufferCPU, blackmagicRawResourceUsageReadCPUWriteCPU, &m_bitStream));
			m_bitStreamSizeBytes = bitStreamSizeBytes;
		}

		// Create the job
		VERIFY(clipEx->CreateJobReadFrame(frameIndex, m_bitStream, bitStreamSizeBytes, jobRead));
	}

	void createDecodeJob(IBlackmagicRawJob** decodeJob)
	{
		uint32_t decodedBufferSizeBytes;
		VERIFY(m_manualDecoder->GetDecodedSizeBytes(m_frameState, &decodedBufferSizeBytes));

		// Allocate the buffers if necessary
		if (decodedBufferSizeBytes > m_decodedBufferSizeBytes)
		{
			if (m_decodedBufferCPU != nullptr)
			{
				VERIFY(m_resourceManager->ReleaseResource(nullptr, nullptr, m_decodedBufferCPU, blackmagicRawResourceTypeBufferCPU));
			}

			if (m_decodedBufferGPU != nullptr)
			{
				VERIFY(m_resourceManager->ReleaseResource(m_context, m_commandQueue, m_decodedBufferGPU, m_gpuResourceType));
			}

			VERIFY(m_resourceManager->CreateResource(nullptr, nullptr, decodedBufferSizeBytes, blackmagicRawResourceTypeBufferCPU, blackmagicRawResourceUsageReadCPUWriteCPU, &m_decodedBufferCPU));
			VERIFY(m_resourceManager->CreateResource(m_context, m_commandQueue, decodedBufferSizeBytes, m_gpuResourceType, blackmagicRawResourceUsageReadGPUWriteGPU, &m_decodedBufferGPU));

			m_decodedBufferSizeBytes = decodedBufferSizeBytes;
		}

		// Create the job
		VERIFY(m_manualDecoder->CreateJobDecode(m_frameState, m_bitStream, m_decodedBufferCPU, decodeJob));
	}

	void createProcessJob(IBlackmagicRawJob** processJob)
	{
		uint32_t processedBufferSizeBytes;
		VERIFY(m_manualDecoder->GetProcessedSizeBytes(m_frameState, &processedBufferSizeBytes));

		// Allocate the Processed Buffer if necessary
		if (processedBufferSizeBytes > m_processedBufferSizeBytes)
		{
			if (m_processedBuffer != nullptr)
			{
				VERIFY(m_resourceManager->ReleaseResource(nullptr, nullptr, m_processedBuffer, blackmagicRawResourceTypeBufferCPU));
			}

			VERIFY(m_resourceManager->CreateResource(m_context, m_commandQueue, processedBufferSizeBytes, m_gpuResourceType, blackmagicRawResourceUsageReadCPUWriteGPU, &m_processedBuffer));
			m_processedBufferSizeBytes = processedBufferSizeBytes;
		}

		// Copy the Decoded buffer from main memory to the GPU
		uint32_t decodedBufferSizeBytes;
		VERIFY(m_manualDecoder->GetDecodedSizeBytes(m_frameState, &decodedBufferSizeBytes));
		VERIFY(m_resourceManager->CopyResource(m_context, m_commandQueue, m_decodedBufferCPU, blackmagicRawResourceTypeBufferCPU, m_decodedBufferGPU, m_gpuResourceType, decodedBufferSizeBytes, true));

		// Create the job
		VERIFY(m_manualDecoder->CreateJobProcess(m_context, m_commandQueue, m_frameState, m_decodedBufferGPU, nullptr, m_processedBuffer, m_post3DLUTBufferGPU, processJob));
	}

private:
	IBlackmagicRawResourceManager*		m_resourceManager			= nullptr;
	IBlackmagicRawManualDecoderFlow2*	m_manualDecoder				= nullptr;
	void*								m_post3DLUTBufferGPU		= nullptr;
	void*								m_context					= nullptr;
	void*								m_commandQueue				= nullptr;
	BlackmagicRawResourceType			m_gpuResourceType			= blackmagicRawResourceTypeBufferCPU;

	uint32_t							m_bitStreamSizeBytes		= 0;
	void*								m_bitStream					= nullptr;

	uint32_t							m_frameStateSizeBytes		= 0;
	void*								m_frameState				= nullptr;

	uint32_t							m_decodedBufferSizeBytes	= 0;
	void*								m_decodedBufferCPU			= nullptr;
	void*								m_decodedBufferGPU			= nullptr;

	uint32_t							m_processedBufferSizeBytes	= 0;
	void*								m_processedBuffer			= nullptr;
};

struct UserData
{
	unsigned long long m_frameIndex;
	BufferManagerFlow2* m_bufferManager;
};

class CameraCodecCallback : public IBlackmagicRawCallback
{
public:
	explicit CameraCodecCallback() = default;
	virtual ~CameraCodecCallback() = default;

	virtual void ReadComplete(IBlackmagicRawJob* readJob, HRESULT result, IBlackmagicRawFrame* frame)
	{
		UserData* userData = nullptr;
		VERIFY(readJob->GetUserData((void**)&userData));
		BufferManagerFlow2* bufferManager = userData->m_bufferManager;
		IBlackmagicRawJob* decodeJob = nullptr;

		if (result == S_OK)
		{
			VERIFY(frame->SetResourceFormat(s_resourceFormat));
			bufferManager->populateFrameStateBuffer(frame);
			bufferManager->createDecodeJob(&decodeJob);
			decodeJob->SetUserData(userData);
			result = decodeJob->Submit();
		}

		if (result != S_OK)
		{
			if (decodeJob)
				decodeJob->Release();
		}

		// Release the job (The API holds a reference to the job, so this call is just transferring ownership)
		VERIFY(decodeJob->Release());
	}

	virtual void DecodeComplete(IBlackmagicRawJob* decodeJob, HRESULT result)
	{
		UserData* userData = nullptr;
		VERIFY(decodeJob->GetUserData((void**)&userData));
		BufferManagerFlow2* bufferManager = userData->m_bufferManager;

		if (! SUCCEEDED(result))
		{
			std::lock_guard<std::mutex> lock(m_mutex);

			delete userData;
			return;
		}

		// Start the Process job
		IBlackmagicRawJob* processJob = nullptr;
		bufferManager->createProcessJob(&processJob);
		VERIFY(processJob->SetUserData(userData));
		VERIFY(processJob->Submit());

		// Release the job (The API holds a reference to the job, so this call is just transferring ownership)
		VERIFY(processJob->Release());
	}

	virtual void ProcessComplete(IBlackmagicRawJob* job, HRESULT result, IBlackmagicRawProcessedImage* processedImage)
	{
		std::lock_guard<std::mutex> scopedLock(m_mutex);
		if (result == S_OK)
		{
			processedImage->AddRef();
			m_processedImage = processedImage;
		}

		m_finished = true;
	}

	IBlackmagicRawProcessedImage* takeProcessedImage()
	{
		std::lock_guard<std::mutex> scopedLock(m_mutex);
		IBlackmagicRawProcessedImage* processedImage = m_processedImage;
		m_processedImage = nullptr;
		m_finished = false;
		return processedImage;
	}

	bool decodeFinished()
	{
		std::lock_guard<std::mutex> scopedLock(m_mutex);
		return m_finished;
	}

	virtual void TrimProgress(IBlackmagicRawJob*, float) {}
	virtual void TrimComplete(IBlackmagicRawJob*, HRESULT) {}
	virtual void SidecarMetadataParseWarning(IBlackmagicRawClip*, CFStringRef, uint32_t, CFStringRef) {}
	virtual void SidecarMetadataParseError(IBlackmagicRawClip*, CFStringRef, uint32_t, CFStringRef) {}
	virtual void PreparePipelineComplete(void*, HRESULT) {}

	virtual HRESULT STDMETHODCALLTYPE QueryInterface(REFIID, LPVOID*)
	{
		return E_NOTIMPL;
	}

	virtual ULONG STDMETHODCALLTYPE AddRef(void)
	{
		return 0;
	}

	virtual ULONG STDMETHODCALLTYPE Release(void)
	{
		return 0;
	}
private:
	IBlackmagicRawProcessedImage*	m_processedImage	= nullptr;
	std::mutex						m_mutex;
	bool m_finished = false;
};


@interface ViewController ()
@property (nonatomic, strong) UIImageView* imgView;
@end

@implementation ViewController

- (void)viewDidLoad {
	[super viewDidLoad];
	// Do any additional setup after loading the view.
	NSBundle* mainBundle = [NSBundle mainBundle];
	NSURL* clipURL = nullptr;

	IBlackmagicRawFactory* factory = nullptr;
	IBlackmagicRaw* codec = nullptr;
	IBlackmagicRawConfigurationEx* configurationEx = nullptr;
	IBlackmagicRawResourceManager* resourceManager = nullptr;
	IBlackmagicRawManualDecoderFlow2* manualDecoder = nullptr;

	IBlackmagicRawClip* clip = nullptr;
	IBlackmagicRawClipEx* clipEx = nullptr;
	IBlackmagicRawClipProcessingAttributes* clipProcessingAttributes = nullptr;
	IBlackmagicRawPost3DLUT* clipPost3DLUT = nullptr;
	void* clipPost3DLUTBufferGPU = nullptr;
	IBlackmagicRawJob* readJob = nullptr;

	BlackmagicRawPipeline pipeline;
	BlackmagicRawResourceType gpuResourceType = blackmagicRawResourceTypeBufferMetal;
	void* context = nullptr;
	void* commandQueue = nullptr;

	CameraCodecCallback callback;
	unsigned long long frameIndex = 0;

	factory = CreateBlackmagicRawFactoryInstance();
	assert(factory != nullptr);

	VERIFY(factory->CreateCodec(&codec));

	IBlackmagicRawPipelineDeviceIterator* deviceIterator = nullptr;
	VERIFY(factory->CreatePipelineDeviceIterator(blackmagicRawPipelineMetal, blackmagicRawInteropNone, &deviceIterator));
	IBlackmagicRawPipelineDevice* device = nullptr;
	VERIFY(deviceIterator->CreateDevice(&device));

	VERIFY(device->GetPipeline(&pipeline, &context, &commandQueue));

	IBlackmagicRawConfiguration* config = nullptr;
	VERIFY(codec->QueryInterface(IID_IBlackmagicRawConfiguration, (void**)&config));

	VERIFY(config->SetPipeline(blackmagicRawPipelineMetal, context, commandQueue));

	VERIFY(codec->QueryInterface(IID_IBlackmagicRawConfigurationEx, (void**)&configurationEx));

	VERIFY(configurationEx->GetResourceManager(&resourceManager));
	clipURL = [mainBundle URLForResource:@"sample" withExtension:@"braw"];
	assert (clipURL != nullptr);

	CFStringRef clipPath = CFURLCopyFileSystemPath((__bridge CFURLRef)clipURL.absoluteURL, kCFURLPOSIXPathStyle);
	VERIFY(codec->OpenClip(clipPath, &clip));

	VERIFY(codec->SetCallback(&callback));

	VERIFY(codec->QueryInterface(IID_IBlackmagicRawManualDecoderFlow2, (void**)&manualDecoder));

	VERIFY(clip->QueryInterface(IID_IBlackmagicRawClipEx, (void**)&clipEx));
	VERIFY(clip->QueryInterface(IID_IBlackmagicRawClipProcessingAttributes, (void**)&clipProcessingAttributes));
	VERIFY(clipProcessingAttributes->GetPost3DLUT(&clipPost3DLUT));
	if (clipPost3DLUT != nullptr)
	{
		BlackmagicRawResourceType post3dLutResourceType;
		VERIFY(clipPost3DLUT->GetResourceGPU(context, commandQueue, &post3dLutResourceType, &clipPost3DLUTBufferGPU));

		assert(post3dLutResourceType == gpuResourceType);
	}


	BufferManagerFlow2 bufferManager(resourceManager, manualDecoder, clipPost3DLUTBufferGPU, context, commandQueue, blackmagicRawResourceTypeBufferMetal);

	bufferManager.createReadJob(clipEx, frameIndex, &readJob);

	// Allocate and populate the UserData struct
	UserData* userData = new UserData();
	userData->m_bufferManager = &bufferManager;
	userData->m_frameIndex = frameIndex;
	VERIFY(readJob->SetUserData(userData));

	// Start the Read job
	VERIFY(readJob->Submit());

	// Release the job (The API holds a reference to the job, so this call is just transferring ownership)
	VERIFY(readJob->Release());

	while (! callback.decodeFinished())
	{
		std::this_thread::sleep_for(std::chrono::milliseconds(1));
	}

	IBlackmagicRawProcessedImage* processedImage = callback.takeProcessedImage();

	unsigned int width = 0;
	unsigned int height = 0;
	unsigned int sizeBytes = 0;
	void* GPUimageData = nullptr;

	VERIFY(processedImage->GetWidth(&width));

	VERIFY(processedImage->GetHeight(&height));

	VERIFY(processedImage->GetResourceSizeBytes(&sizeBytes));

	VERIFY(processedImage->GetResource(&GPUimageData));

	uint32_t bytesPerRow = width * 4;

	BlackmagicRawResourceType resourceType;
	VERIFY(processedImage->GetResourceType(&resourceType));

	assert (resourceType == blackmagicRawResourceTypeBufferMetal);
	id<MTLBuffer> metalBuffer = (__bridge id<MTLBuffer>)GPUimageData;
	void* imageData = metalBuffer.contents;

	uint32_t pixelDataSizeBytes;
	processedImage->GetResourceSizeBytes(&pixelDataSizeBytes);

	CFDataRef pixelData = CFDataCreate(kCFAllocatorDefault, reinterpret_cast<const UInt8 *>(imageData), sizeBytes);

	CGDataProviderRef pixelDataProvider = CGDataProviderCreateWithCFData(pixelData);
	CGColorSpaceRef colourSpace = CGColorSpaceCreateDeviceRGB();

	CGImageRef brawImage = CGImageCreate(width, height, 8, 32, bytesPerRow, colourSpace, kCGBitmapByteOrderDefault | kCGImageAlphaNoneSkipLast, pixelDataProvider, nullptr, true, kCGRenderingIntentDefault);

	UIImage* img = [UIImage imageWithCGImage:brawImage];

	self.imgView = [[UIImageView alloc] initWithFrame:self.view.bounds];
	self.imgView.contentMode = UIViewContentModeScaleAspectFit;
	self.imgView.image = img;
	self.imgView.center = self.view.center;
	[self.view addSubview:self.imgView];

	codec->FlushJobs();

	if (clip != nullptr)
		clip->Release();

	if (codec != nullptr)
		codec->Release();

	if (factory != nullptr)
		factory->Release();
}

@end
